//------------------------------------------------------------------------------
// File:               para-equalizer-x16-stereo.cpp
// Standalone Plugin:  Parametrischer Entzerrer x16 Stereo - Parametric Equalizer x16 Stereo [STANDALONE]
// Standalone UID:     'para_equalizer_x16_stereo'
// Version:            1.0.38
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define STANDALONE_PLUGIN_UID  "para_equalizer_x16_stereo"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_STANDALONE_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/standalone/main.h>
#undef LSP_PLUG_IN_STANDALONE_MAIN_IMPL
