//------------------------------------------------------------------------------
// File:               mixer-x8-stereo.cpp
// Standalone Plugin:  Mischer x8 Stereo - Mixer x8 Stereo [STANDALONE]
// Standalone UID:     'mixer_x8_stereo'
// Version:            1.0.25
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define STANDALONE_PLUGIN_UID  "mixer_x8_stereo"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_STANDALONE_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/standalone/main.h>
#undef LSP_PLUG_IN_STANDALONE_MAIN_IMPL
