//------------------------------------------------------------------------------
// File:               sc-gate-lr.cpp
// Standalone Plugin:  Sidechain-Gate LeftRight - Sidechain Gate LeftRight [STANDALONE]
// Standalone UID:     'sc_gate_lr'
// Version:            1.0.37
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define STANDALONE_PLUGIN_UID  "sc_gate_lr"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_STANDALONE_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/standalone/main.h>
#undef LSP_PLUG_IN_STANDALONE_MAIN_IMPL
