//------------------------------------------------------------------------------
// File:            comp-delay-mono.cpp
// VST2 Plugin:     Verzögerungsausgleicher Mono - Delay Compensator Mono [VST2]
// VST2 UID:        'jav8'
// Version:         1.0.37
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define VST2_PLUGIN_UID     "jav8"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_VST2_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/vst2/main.h>
#undef LSP_PLUG_IN_VST2_MAIN_IMPL
