//------------------------------------------------------------------------------
// File:            beat-breather-stereo.cpp
// VST2 Plugin:     Beat Breather Stereo - Beat Breather Stereo [VST2]
// VST2 UID:        'bb1s'
// Version:         1.0.22
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define VST2_PLUGIN_UID     "bb1s"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_VST2_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/vst2/main.h>
#undef LSP_PLUG_IN_VST2_MAIN_IMPL
