//------------------------------------------------------------------------------
// File:               oscillator-mono.cpp
// Standalone Plugin:  Oszillator Mono - Oscillator Mono [STANDALONE]
// Standalone UID:     'oscillator_mono'
// Version:            1.0.33
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define STANDALONE_PLUGIN_UID  "oscillator_mono"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_STANDALONE_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/standalone/main.h>
#undef LSP_PLUG_IN_STANDALONE_MAIN_IMPL
