//------------------------------------------------------------------------------
// File:               sc-mb-dyna-processor-lr.cpp
// Standalone Plugin:  Sidechain Multi-band Dynamikprozessor LeftRight x8 - Sidechain Multiband Dynamics Processor LeftRight x8 [STANDALONE]
// Standalone UID:     'sc_mb_dyna_processor_lr'
// Version:            1.0.30
//------------------------------------------------------------------------------

// Pass Plugin UID for factory function
#define STANDALONE_PLUGIN_UID  "sc_mb_dyna_processor_lr"

#include <lsp-plug.in/common/types.h>

// Include factory function implementation
#define LSP_PLUG_IN_STANDALONE_MAIN_IMPL
    #include <lsp-plug.in/plug-fw/wrap/standalone/main.h>
#undef LSP_PLUG_IN_STANDALONE_MAIN_IMPL
